#!/usr/bin/ruby
# encoding: UTF-8

# == Synopsis 
#   This program processes an Saleae logic analyzer SPI export
#   file to generate a more readable listing by consolidating
#   packet information to a single row.  Optionally repeated
#   packets can be collapsed to a single row plus count.
#
# == Examples
#   spi_format myspidata.csv myspidata_fmt.csv
#
# == Usage 
#   spi_format [options] source_file output_file
#
#   For help use: spi_format -h
#
# == Author
#   hexfet
#
# == Licensed under GNU GPL version 3


# Process command line options
require 'optparse'
require 'ostruct'
require 'date'
require 'csv'

# These column names are for Saleae SPI export file
COL_TIME = 'Time [s]'
COL_PID = 'Packet ID'
COL_MOSI = 'MOSI'
COL_MISO = 'MISO'
HEADERS = [COL_TIME, COL_PID, COL_MOSI, COL_MISO]

# additional headers in output file
COL_DIFF = 'Diff [ms]'
COL_BYTES = 'Bytes'
COL_REPEAT = 'Repeat'
OUTHEADERS = [COL_TIME, COL_DIFF, COL_PID, COL_MOSI, COL_MISO, COL_BYTES, COL_REPEAT]

    

class App
  attr_reader :options
  
  def initialize(arguments, stdin)
    @arguments = arguments

    # Set defaults
    @options = OpenStruct.new
    @options.collapse = true
    @options.mosi = nil
    @options.miso = nil
  end

  # Parse options, check arguments, then process the command
  def run
    @start_time = DateTime.now

    if parsed_options? && arguments_valid? 

      process_arguments            

      process_command
      
      @stop_time = DateTime.now
      puts "Elapsed time #{TimeFormatter.elapsed((86400*(@stop_time-@start_time)).to_f)}"
      
    else
      output_usage
    end
  end
  
  protected

    def parsed_options?
      # Specify options
      opts = OptionParser.new 
      opts.banner = "Usage: spi_format [options] csvfile outputfile (csvfile is Saleae SPI export format)"
      opts.on_head(nil, '--nocollapse', 'Do not collapse repeated packets') { @options.collapse = false }
      opts.on_head('-m', '--mosi DATA', 'Only output packets where MOSI matches DATA (separate multiple values with comma)') {|data| @options.mosi = data.split(',') }
      opts.on_head('-i', '--miso DATA', 'Only output packets where MISO matches DATA (separate multiple values with comma)') {|data| @options.miso = data.split(',') }
      opts.on_tail('-h', '--help')    { output_help }                          
      @helpstring = opts.to_s
            
      begin
        opts.parse!(@arguments)
      rescue Exception => e
        puts e.to_s
        return false
      end
      process_options
      true      
    end

    # Performs post-parse processing on options
    def process_options
    end
    
    def output_options
      puts "Options:\n"
      @options.marshal_dump.each do |name, val|        
        puts "  #{name} = #{val}" unless name == :password
      end
    end

    # True if required arguments were provided
    def arguments_valid?
      true if @arguments.length >= 1 
    end
    
    # Setup the arguments
    def process_arguments

    end
    
    def output_help
      output_usage
    end
    
    def output_usage
      puts @helpstring
      exit 1
    end
    
    def packet_write(outputfile, outpacket, prev_time)
      # prettify the output
      outpacket[COL_DIFF] = "%10.3f" % ((outpacket[COL_TIME].to_f - prev_time) * 1000.0)
      outpacket[COL_MOSI] = outpacket[COL_MOSI].gsub(/(..)/, '\1 ')
      outpacket[COL_MISO] = outpacket[COL_MISO].gsub(/(..)/, '\1 ')
      outpacket[COL_REPEAT] = nil if outpacket[COL_REPEAT] == 0
      outputfile << outpacket
    end

    def process_command
      begin
        packet_reader = PacketReader.new @arguments[0], @options.mosi, @options.miso
      rescue Exception => e
        puts  "Cannot find input file #{@arguments[0]}\n\n"
        exit 1
      end
      puts "Processing input file #{@arguments[0]}"

      # open output file
      begin
        outputfile = CSV.open(@arguments[1], 'w', headers: true, write_headers: true)
      rescue Exception => e
        puts "Cannot open output file #{@arguments[1]}\n\n"
        exit 1
      end
      puts "Opened output file #{@arguments[1]}"

      outputfile << CSV::Row.new(OUTHEADERS, OUTHEADERS, true)

      printf("Rows Processed\n")
      packet = nil
      outpacket = nil
      prev_time = nil
      packets_out = 0

      loop do
        packet = packet_reader.next
        break unless packet
        prev_time ||= packet[COL_TIME].to_f

        if @options.collapse and (outpacket && (outpacket[COL_MOSI] + outpacket[COL_MISO]) == (packet[COL_MOSI] + packet[COL_MISO]))
          outpacket[COL_REPEAT] += 1
        else
          if outpacket
            packet_write outputfile, outpacket, prev_time
            prev_time = outpacket[COL_TIME].to_f
            packets_out += 1
          end
          outpacket = packet
        end
        print "#{packet_reader.packets_in}\r" if packet_reader.packets_in % 7 == 0
      end
      packet_write outputfile, outpacket, prev_time if outpacket
      packets_out += 1
      puts packets_out

      packet_reader.close
      outputfile.close

      puts "\nCompleted"
      puts "Packets read      : #{packet_reader.packets_in}"
      puts "Packets filtered  : #{packet_reader.packets_filtered}"
      puts "Packets out       : #{packets_out}"
      puts "Average rate      : " + format("%-8.3f", packet_reader.packets_in.to_f/(86400*(DateTime.now-@start_time))) + " rows/second"
      puts "\n"
    end
end

class PacketReader
  attr_reader :packets_in, :packets_filtered

  def initialize(filename, mosi_filter=nil, miso_filter=nil)
    @csvsource = CSV.open(filename, headers: true)
    @row = nil
    @packets_in = 0
    @packets_filtered = 0
    @mosi_filter = mosi_filter
    @miso_filter = miso_filter
  end

  def close
    @csvsource.close
  end

  def conv_hex(field)
    return ("%02X" % field.gsub("'","").to_i) if field[0] == "'"
    return ("%02X" % field.ord) if field[0] == "\\"
    "%02X" % field.hex
  rescue Exception => e
    "??"
  end

  def next
    packet = nil
    loop do
      packet = new_packet @row || @csvsource.first
      break unless packet
      @row = @csvsource.first
      while @row and packet[COL_PID] == @row[COL_PID] do
        packet[COL_MOSI] << conv_hex(@row[COL_MOSI])
        packet[COL_MISO] << conv_hex(@row[COL_MISO])
        packet[COL_BYTES] += 1
        @row = @csvsource.first
      end
      @packets_in += 1

      break unless @mosi_filter && !@mosi_filter.detect {|mosi| mosi == packet[COL_MOSI][0, mosi.length]} \
                or @miso_filter && !@miso_filter.detect {|miso| miso == packet[COL_MISO][0, miso.length]}

      @packets_filtered += 1
    end
    packet
  end

  def new_packet(row)
    return unless row
    CSV::Row.new(OUTHEADERS, ["%10.6f" % row[COL_TIME], 0, row[COL_PID], \
        conv_hex(row[COL_MOSI]), conv_hex(row[COL_MISO]), 1, 0])
  end
end


class TimeFormatter
  def self.elapsed(time_elapsed)
    hours = time_elapsed.to_i / 3600
    minutes = (time_elapsed.to_i / 60) % 60
    seconds = time_elapsed % 60
    return hours.to_s + ":" + format("%02d", minutes) + ":" + format("%06.3f", seconds)
  end
end


# Create and run the application
app = App.new(ARGV, STDIN)
app.run

